from collections.abc import Callable
from typing import TypeVar, get_origin, get_type_hints, overload

from ._event import Event, event


__all__ = [
    "events",
]


_T = TypeVar("_T", bound=type)


@overload
def events(cls: _T, /) -> _T: ...


@overload
def events(*, prefix: str = "", collection: str | None = None) -> Callable[[_T], _T]: ...


def events(cls: _T | None = None, /, *, prefix: str = "", collection: str | None = None) -> _T | Callable[[_T], _T]:
    """
    Adds event fields or properties based on the annotations defined in the class and the specified prefix.

    If this decorator is used to add event fields (``collection=None``),
    then field annotations of type ``Event`` are turned into field assignments in a generated ``__init__`` method.
    The ``__init__`` method will call the original ``__init__`` method at the end of its body if defined,
    or call ``super().__init__`` at the beginning of its body otherwise.

    Example::

        @events(prefix="on_")
        class EventFieldsExample:
            on_change: Event[object, str]
            on_input: Event[int]
            ignored: Event[[]]

    is equivalent to::

        class EventFieldsExample:
            def __init__(self, *args, **kwargs) -> None:
                super().__init__(*args, **kwargs)
                self.on_change: Event[object, str] = Event()
                self.on_input: Event[int] = Event()

    Event properties require ``collection`` to be set to the attribute name of an ``EventHandlerCollection``.
    Field annotations of type ``event`` (NOT ``Event``) are turned into event property definitions,
    using the collection attribute provided and the field name without ``prefix`` as the key.

    Example::

        @events(collection="__events", prefix="on_")
        class EventPropertiesExample:
            on_change: event[object, str]
            on_input: event[int]

            def __init__(self) -> None:
                self.__events = EventHandlerList()

    is equivalent to::

        class EventPropertiesExample:
            @event[object, str]
            def on_change():
                def add(self, value):
                    self.__events.add_handler("change", value)
                def remove(self, value):
                    self.__events.remove_handler("change", value)
                return add, remove

            @event[int]
            def on_input():
                def add(self, value):
                    self.__events.add_handler("input", value)
                def remove(self, value):
                    self.__events.remove_handler("input", value)
                return add, remove

            def __init__(self) -> None:
                self.__events = EventHandlerList()

    If your class defines a large number of events, the storage cost of one field per event might not be acceptable.
    For those situations, event properties can be used to store the events in another data structure.

    Event properties are slower than event fields due to the additional data structure.
    The trade-off is between memory and speed.
    If your class defines many events that are infrequently raised, you'll want to use event properties.

    Args:
     - cls (type): A class with event annotations.
     - prefix (str, optional): Only process annotations that start with the prefix. Defaults to "".
     - collection (str | None, optional): Attribute name of an EventHandlerCollection or None. Defaults to None.

    Returns:
        (type): cls
    """

    if cls is None:
        if collection:  # treat "" as None as well
            return lambda cls: _create_events(cls, prefix, collection)
        return lambda cls: _create_init(cls, prefix)

    if collection:
        return _create_events(cls, prefix, collection)
    return _create_init(cls, prefix)


def _create_init(cls: _T, prefix: str, /) -> _T:
    body_lines: list[str] = []

    for (name, T) in get_type_hints(cls).items():
        if name.startswith(prefix):
            if (get_origin(T) or T) is Event:
                body_lines.append(f"self.{name} = Event()")

    if not body_lines:
        return cls

    globals: dict[str, object] = {"Event": Event}
    locals: dict[str, Callable[..., None]] = {}

    replace = "__init__" in vars(cls)

    if replace:
        globals["__init__"] = cls.__init__
        body_lines.append("__init__(self, *args, **kwargs)")
    else:
        globals["__class__"] = cls
        body_lines.insert(0, "super(__class__, self).__init__(*args, **kwargs)")

    body = "\n".join(f"\t{line}" for line in body_lines)
    exec(f"def __init__(self, /, *args, **kwargs) -> None:\n{body}", globals, locals)
    func = locals["__init__"]

    if replace:
        for attr in (
            "__module__",
            "__name__",
            "__qualname__",
            "__doc__",
            "__annotations__",
            "__dict__"
        ):
            try:
                value = getattr(cls.__init__, attr)
            except AttributeError:
                pass
            else:
                setattr(func, attr, value)
    else:
        func.__module__ = cls.__module__
        func.__name__ = "__init__"
        func.__qualname__ = f"{cls.__qualname__}.__init__"

    cls.__init__ = func
    return cls


def _create_events(cls: _T, prefix: str, collection: str, /) -> _T:
    # Assume collection is a valid identifier
    if len(collection) > 2 and collection[:2] == "__" and collection[-2:] != "__":
        collection = f"_{cls.__name__}{collection}"

    locals = {}
    exec(
        "def f(name, /):\n"
        "  def add(self, value, /):\n"
        f"    self.{collection}.add_handler(name, value)\n"

        "  def remove(self, value, /):\n"
        f"    self.{collection}.remove_handler(name, value)\n"

        "  return event(lambda: (add, remove))\n",
        {"event": event},
        locals
    )
    create_event: Callable[[str], event[...]] = locals["f"]

    for (name, T) in get_type_hints(cls).items():
        if name.startswith(prefix):
            if (get_origin(T) or T) is event:
                setattr(cls, name, create_event(name.removeprefix(prefix)))
    return cls



import sys
from collections.abc import Callable, Collection, Iterator
from typing import Any, Generic, ParamSpec, final


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


__all__ = [
    "accessors",
    "event",
    "Event",
    "EventHandler",
]

# Python does not provide a void type, which is a useful feature in callbacks,
# and is clearly different from None.
# Comparison with TS:
#     function foo(): void {} <==> def foo() -> None: pass
#     let bar: () => void;    <==> bar: Callable[[], Any]
void = None | Any

P = ParamSpec("P")
EventHandler = Callable[P, void]
accessors = tuple[Callable[[Any, EventHandler[P]], void], Callable[[Any, EventHandler[P]], void]]


@final
class Event(Collection[EventHandler[P]]):
    """
    Represents an event delegate that handlers can subscribe to.

    The type argument specifies the event data parameters::

        event = Event[str, int]()
        # accepts handlers (str, int) -> void

    Handlers can subscribe to and unsubscribe from an event by::

        event += event_handler
        event -= event_handler

    An event can be raised by invoking itself with the necessary arguments::

        event(...)

    Type Args:
     - P (ParamSpec): Event data parameter specification.
    """

    __slots__ = ("__handlers")

    __handlers: list[EventHandler[P]]

    def __init__(self, *handlers: EventHandler[P]) -> None:
        """
        Initializes a new instance of the ``Event`` class.

        Args:
         - *handlers ((**P) -> void): List of handlers to subscribe to the event.
        """

        self.__handlers = [*handlers]

    def __iadd__(self, handler: EventHandler[P], /) -> Self:
        """
        Subscribes the handler to this event.

        Args:
         - handler ((**P) -> void): An event handler.

        Returns:
            (Self): This event.
        """

        self.__handlers.append(handler)
        return self

    def __isub__(self, handler: EventHandler[P], /) -> Self:
        """
        Unsubscribes the handler from this event.

        If the handler has been added multiple times, removes only the last occurrence.

        Args:
         - handler ((**P) -> void): An event handler

        Returns:
            (Self): This event
        """

        for i in reversed(range(len(self.__handlers))):
            if self.__handlers[i] is handler:
                del self.__handlers[i]
                break
        return self

    def __contains__(self, handler: object, /) -> bool:
        """
        Returns whether the handler has been subscribed to this event.

        Args:
         - handler (object): An event handler.

        Returns:
            (bool): True if handler is subscribed, False otherwise.
        """

        return self.__handlers.__contains__(handler)

    def __iter__(self) -> Iterator[EventHandler[P]]:
        """
        Returns an iterator from the list of subscribed handlers.

        Yields:
            ((**P) -> void): A subscribed event handler.
        """

        return self.__handlers.__iter__()

    def __len__(self) -> int:
        """
        Returns the number of subscribed handlers.

        Returns:
            (int): The number of subscribed event handlers.
        """

        return self.__handlers.__len__()

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """
        Raises this event.

        Handlers will be invoked in the order they subscribed.
        """

        for handler in [*self.__handlers]:  # apparently faster than list.copy(), list[:] or even (*list, )
            handler(*args, **kwargs)


@final
class event(Generic[P]):
    """
    A decorator used to declare an event property in a publisher class.

    The definition function must return a tuple of accessors ``(add, remove)``
    each of type ``(Self, (**P) -> void) -> void``.

    The subscribe and unsubscribe operators::

        obj.event += handler
        obj.event -= handler

    will invoke the respective `add` and `remove` accessors with the arguments
    `(obj, handler)`.

    An event property is not bound to an instance object, thus it must be accessed as
    an attribute of an object in order for the accessors to be invoked.
    The following code::

        e = obj.event
        e += handler  # no effect on obj.event

    has no effect on `obj.event` and will not call the `add` accessor,
    since the argument `self` cannot be supplied.

    Unlike event fields, an event property cannot be invoked directly.
    Use the underlying event that the accessors point to instead.

    Example::

        class Example:
            # Infer event type from the accessor parameter type
            @event
            def item_added():
                def add(self: Self, value: EventHandler[int, object]):
                    ...
                def remove(self: Self, value: EventHandler[int, object]):
                    ...
                return add, remove

            # Infer event type from the accessors type
            @event
            def item_removed() -> accessors[int, object]:
                def add(self: Self, value):
                    ...
                def remove(self: Self, value):
                    ...
                return add, remove

            # Explicit event type
            @event[int, object]
            def item_changed():
                def add(self: Self, value):
                    ...
                def remove(self: Self, value):
                    ...
                return add, remove
    """

    __slots__ = ("__add", "__remove")

    def __init__(self, f: Callable[[], accessors[P]], /) -> None:
        """
        Initializes a new instance of the ``event`` class.

        Args:
            f (() -> accessors[P]): An event definition.
        """
        (self.__add, self.__remove) = f()

    def __get__(self, instance: object, cls: type, /) -> Self:
        return self

    def __set__(self, instance: object, value: tuple[EventHandler[P], bool], /) -> None:
        (handler, add) = value
        if add:
            self.__add(instance, handler)
        else:
            self.__remove(instance, handler)

    def __iadd__(self, handler: EventHandler[P], /) -> tuple[EventHandler[P], bool]:
        return (handler, True)

    def __isub__(self, handler: EventHandler[P], /) -> tuple[EventHandler[P], bool]:
        return (handler, False)



from unittest.mock import Mock

import pytest

from cs_events import Event, EventHandlerCollection, EventHandlerList, event, events


def test_events_field() -> None:
    @events
    class TestClass:
        event1: Event[[]]

    assert isinstance(TestClass().event1, Event)


def test_events_field_init() -> None:
    @events
    class TestClass0:
        no_events: None

    assert "__init__" not in vars(TestClass0)

    __init__ = Mock()

    @events
    class TestClass1:
        event1: Event[[]]

        def __init__(self) -> None:
            __init__(self)

    obj1 = TestClass1()
    __init__.assert_called_once_with(obj1)
    assert isinstance(obj1.event1, Event)

    @events
    class TestClass2(TestClass1):
        event2: Event[[]]

    __init__.reset_mock()
    obj2 = TestClass2()
    __init__.assert_called_once_with(obj2)
    assert isinstance(obj2.event1, Event)
    assert isinstance(obj2.event2, Event)


def test_events_field_prefix() -> None:
    @events(prefix="on_")
    class TestClass:
        event1: Event[[]]
        on_event2: Event[[]]

    obj = TestClass()

    with pytest.raises(AttributeError, match="'TestClass' object has no attribute 'event1'"):
        obj.event1
    assert isinstance(obj.on_event2, Event)


def test_events_properties() -> None:
    @events(collection="__events", prefix="event")
    class TestClass:
        event1: event[int]
        event20: event[str, str]
        event3: event[...]

        def __init__(self) -> None:
            self.__events = EventHandlerList()

        @property
        def events(self) -> EventHandlerCollection:
            return self.__events

    assert isinstance(TestClass.event1, event)
    assert isinstance(TestClass.event20, event)
    assert isinstance(TestClass.event3, event)

    obj = TestClass()

    assert isinstance(obj.event1, event)
    assert isinstance(obj.event20, event)
    assert isinstance(obj.event3, event)

    assert obj.events["1"] is None
    assert obj.events["20"] is None
    assert obj.events["3"] is None

    def event1_handler(_: int) -> None: ...
    def event20_handler(_0: str, _1: str) -> None: ...

    obj.event1 += event1_handler

    assert (e := obj.events["1"]) is not None
    assert len(e) == 1
    assert obj.events["20"] is None
    assert obj.events["3"] is None

    obj.event1 -= event1_handler

    assert (e := obj.events["1"]) is not None
    assert len(e) == 0
    assert obj.events["20"] is None
    assert obj.events["3"] is None

    obj.event20 -= event20_handler

    assert (e := obj.events["1"]) is not None
    assert len(e) == 0
    assert obj.events["20"] is None
    assert obj.events["3"] is None

    obj.event20 += event20_handler

    assert (e := obj.events["1"]) is not None
    assert len(e) == 0
    assert (e := obj.events["20"]) is not None
    assert len(e) == 1
    assert obj.events["3"] is None



import pytest

from cs_events import Event, EventHandlerCollection, EventHandlerDict, EventHandlerList


@pytest.mark.parametrize("collection", [EventHandlerList(), EventHandlerDict()])
def test_EventHandleCollection(collection: EventHandlerCollection) -> None:
    collection = EventHandlerList()
    key1 = object()
    key2 = object()

    def delegate(): ...

    assert collection[key1] is None
    assert collection[key2] is None

    collection.remove_handler(key1, delegate)

    assert collection[key1] is None
    assert collection[key2] is None

    collection.remove_handler(key2, delegate)

    assert collection[key1] is None
    assert collection[key2] is None

    collection.add_handler(key1, delegate)

    assert isinstance(e := collection[key1], Event)
    assert len(e) == 1
    assert collection[key2] is None

    collection.add_handler(key1, delegate)

    assert isinstance(e := collection[key1], Event)
    assert len(e) == 2
    assert collection[key2] is None

    collection.add_handler(key2, delegate)

    assert isinstance(e := collection[key1], Event)
    assert len(e) == 2
    assert isinstance(e := collection[key2], Event)
    assert len(e) == 1



from unittest.mock import Mock

from cs_events import Event, event


def test_Event() -> None:
    def event_handler() -> None: ...

    event = Event()
    assert len(event) == 0

    event = Event(event_handler)
    assert len(event) == 1


def test_Event_impl_Collection() -> None:
    def event_handler1() -> None: ...
    def event_handler2() -> None: ...

    event = Event(
        event_handler1,
        event_handler2,
    )

    assert event_handler1 in event
    assert event_handler2 in event

    assert list(event) == [event_handler1, event_handler2]

    assert len(event) == 2


def test_Event_add_remove() -> None:
    def handler1() -> None: ...
    def handler2() -> None: ...

    event = Event(handler1)
    assert list(event) == [handler1]

    event += handler2
    assert list(event) == [handler1, handler2]

    event -= handler2
    assert list(event) == [handler1]

    event -= handler2
    assert list(event) == [handler1]

    event += handler2
    event += handler1
    event += handler2
    assert list(event) == [handler1, handler2, handler1, handler2]

    event -= handler1
    assert list(event) == [handler1, handler2, handler2]


def test_Event_invoke() -> None:
    handler1 = Mock()
    handler2 = Mock()

    event = Event(handler1, handler2)
    event(12345, "hello", [True, False], a=0, b=None)

    args, kwargs = (12345, "hello", [True, False]), {"a": 0, "b": None}
    handler1.assert_called_once_with(*args, **kwargs)
    handler2.assert_called_once_with(*args, **kwargs)


def test_event() -> None:
    add_mock = Mock()
    remove_mock = Mock()

    class TestClass:
        @event[str]
        def event1():
            return add_mock, remove_mock

    def handler(_: str) -> None: ...

    obj = TestClass()

    add_mock.assert_not_called()
    remove_mock.assert_not_called()

    obj.event1 += handler
    add_mock.assert_called_once_with(obj, handler)

    obj.event1 -= handler
    remove_mock.assert_called_once_with(obj, handler)



from collections.abc import Callable
from typing import Any, Protocol, Union

from ._event import Event


__all__ = [
    "EventHandlerCollection",
    "EventHandlerList",
    "EventHandlerDict",
]


class EventHandlerCollection(Protocol):
    """
    Provides a collection of event handler delegates.
    """

    def __getitem__(self, key: object, /) -> Event[...] | None:
        """
        Gets the event for the specified key.

        Args:
         - key (object): A key to find in the list.

        Returns:
            (Event[...] | None): The event for the specified key, or `None` if an event does not exist.
        """

        ...

    def add_handler(self, key: object, value: Callable[..., Any], /) -> None:
        """
        Adds the delegate to the list.

        Args:
         - key (object): A key that owns the event.
         - value ((...) -> Any): A delegate to add to the list.
        """

        ...

    def remove_handler(self, key: object, value: Callable[..., Any], /) -> None:
        """
        Removes the delegate from the list.

        Args:
         - key (object): A key that owns the event.
         - value ((...) -> Any): A delegate to remove from the list.
        """

        if (e := self[key]) is not None:
            e -= value

    def invoke(self, key: object, /, *args, **kwargs) -> None:
        """
        Invokes an event from the list.

        Equivalent to::

        ```
        e = events[key]
        if e is not None:
            e(...)
        ```

        This method is provided as a shortcut to above,
        since python does not have null-conditional operators::

        ```
        events[key]?.__call__(...)  # C# style func?.Invoke(...)
        events[key]?.(...)  # JavaScript style func?.(...)
        ```

        Args:
         - key (object): A key that owns the event.
        """

        if (e := self[key]) is not None:
            e(*args, **kwargs)


_ListEntry = tuple[object, Event[...], Union["_ListEntry", None]]


class EventHandlerList(EventHandlerCollection):
    """
    Provides a simple linked list of event handler delegates.
    """

    __slots__ = ("__head")

    def __init__(self) -> None:
        """
        Initializes a new instance of the `EventHandlerList` class.
        """

        self.__head: _ListEntry | None = None

    def __getitem__(self, key: object, /) -> Event[...] | None:
        next = self.__head
        while next is not None:
            _key, handler, next = next
            if _key == key:  # use __eq__ instead of "is"
                return handler
        return None

    def add_handler(self, key: object, value: Callable[..., Any], /) -> None:
        if (e := self[key]) is not None:
            e += value
        else:
            self.__head = (key, Event(value), self.__head)


class EventHandlerDict(EventHandlerCollection):
    """
    Provides a simple dictionary of event handler delegates.
    """

    __slots__ = ("__dict")

    def __init__(self) -> None:
        """
        Initializes a new instance of the `EventHandlerDict` class.
        """

        self.__dict: dict[object, Event[...]] = {}

    def __getitem__(self, key: object, /) -> Event[...] | None:
        return self.__dict.get(key)

    def add_handler(self, key: object, value: Callable[..., Any], /) -> None:
        if (e := self[key]) is not None:
            e += value
        else:
            self.__dict[key] = Event(value)



"""
Provides C#-style event handling mechanism.
"""

from ._collections import EventHandlerCollection, EventHandlerDict, EventHandlerList
from ._event import Event, EventHandler, accessors, event
from ._events import events


__all__ = [
    "accessors",
    "event",
    "Event",
    "EventHandler",
    "EventHandlerCollection",
    "EventHandlerDict",
    "EventHandlerList",
    "events",
]
__version__ = "0.2.1"
__author__ = "Daniel Jeong"
